from __future__ import annotations

from typing import Any, Dict, Callable
import re


DEFAULT_SECRET_KEYS = {"api_key", "apikey", "token", "secret", "password", "authorization"}


def redact_dict(d: Dict[str, Any], extra_keys: set[str] | None = None) -> Dict[str, Any]:
    keys = DEFAULT_SECRET_KEYS | (extra_keys or set())
    out: Dict[str, Any] = {}
    for k, v in (d or {}).items():
        if k.lower() in keys:
            out[k] = "***REDACTED***"
        else:
            out[k] = v
    return out


def redact_text(s: str) -> str:
    # minimal: redact bearer tokens
    s = re.sub(r"Bearer\s+[A-Za-z0-9\-\._~\+\/]+=*", "Bearer ***REDACTED***", s)
    return s


Redactor = Callable[[Dict[str, Any]], Dict[str, Any]]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional
from datetime import datetime
from .models import Run, Step


class Store(ABC):
    @abstractmethod
    def init(self) -> None: ...

    @abstractmethod
    def create_run(self, name: str, started_at: datetime, meta: Dict[str, Any]) -> str: ...

    @abstractmethod
    def end_run(self, run_id: str, ended_at: datetime) -> None: ...

    @abstractmethod
    def add_step(
        self,
        run_id: str,
        idx: int,
        kind: str,
        name: str,
        ts: datetime,
        input: Dict[str, Any],
        output: Dict[str, Any],
        error: Optional[Dict[str, Any]],
    ) -> str: ...

    @abstractmethod
    def get_run(self, run_id: str) -> Run: ...

    @abstractmethod
    def list_runs(self, limit: int = 50) -> List[Run]: ...

    @abstractmethod
    def list_steps(self, run_id: str) -> List[Step]: ...



from agentreplay import Recorder, SQLiteStore, diff_runs

def test_diff_detects_change(tmp_path):
    db = tmp_path / "t.db"
    store = SQLiteStore(str(db))

    rec = Recorder(store=store)
    with rec.run("a") as ra:
        rec.tool("t", input={"x": 1}, output={"y": 2})

    rec2 = Recorder(store=store)
    with rec2.run("b") as rb:
        rec2.tool("t", input={"x": 1}, output={"y": 3})

    out = diff_runs(store, ra.id, rb.id)
    assert "DIVERGENCE" in out
    assert "output changed" in out



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional, Literal, List
from datetime import datetime

StepKind = Literal["llm", "tool", "span"]


@dataclass(frozen=True)
class Run:
    id: str
    name: str
    started_at: datetime
    ended_at: Optional[datetime]
    meta: Dict[str, Any]


@dataclass(frozen=True)
class Step:
    id: str
    run_id: str
    idx: int
    kind: StepKind
    name: str
    ts: datetime
    input: Dict[str, Any]
    output: Dict[str, Any]
    error: Optional[Dict[str, Any]]


@dataclass(frozen=True)
class ReplayReport:
    run_id: str
    ok: bool
    steps_replayed: int
    notes: List[str]



from agentreplay import Recorder, SQLiteStore, Replayer

def test_record_and_replay(tmp_path):
    db = tmp_path / "t.db"
    store = SQLiteStore(str(db))
    rec = Recorder(store=store)

    with rec.run("test") as r:
        rec.llm("planner", input={"a": 1}, output={"b": 2})
        rec.tool("tool_x", input={"q": "hi"}, output={"ok": True})

    rep = Replayer(store)
    report = rep.replay(r.id)
    assert report.steps_replayed == 2
    assert report.ok is True



from .recorder import Recorder
from .store_sqlite import SQLiteStore
from .replay import Replayer, ReplayReport
from .diff import diff_runs
from .exporter import export_areplay, import_areplay

__all__ = [
    "Recorder",
    "SQLiteStore",
    "Replayer",
    "ReplayReport",
    "diff_runs",
    "export_areplay",
    "import_areplay",
]



from __future__ import annotations

import json
import zipfile
from datetime import datetime
from typing import Any, Dict, List

from .store import Store
from .models import Run, Step


FORMAT_VERSION = "0.2"


def _json_bytes(obj: Any) -> bytes:
    return json.dumps(obj, ensure_ascii=False, sort_keys=True, indent=2).encode("utf-8")


def export_areplay(store: Store, run_id: str, out_path: str) -> str:
    run: Run = store.get_run(run_id)
    steps: List[Step] = store.list_steps(run_id)

    manifest: Dict[str, Any] = {
        "format": "agentreplay.areplay",
        "version": FORMAT_VERSION,
        "exported_at": datetime.utcnow().isoformat() + "Z",
        "run_id": run_id,
    }

    run_obj = {
        "id": run.id,
        "name": run.name,
        "started_at": run.started_at.isoformat(),
        "ended_at": run.ended_at.isoformat() if run.ended_at else None,
        "meta": run.meta,
    }

    steps_obj = [
        {
            "id": s.id,
            "run_id": s.run_id,
            "idx": s.idx,
            "kind": s.kind,
            "name": s.name,
            "ts": s.ts.isoformat(),
            "input": s.input,
            "output": s.output,
            "error": s.error,
        }
        for s in steps
    ]

    with zipfile.ZipFile(out_path, "w", compression=zipfile.ZIP_DEFLATED) as z:
        z.writestr("manifest.json", _json_bytes(manifest))
        z.writestr("run.json", _json_bytes(run_obj))
        z.writestr("steps.json", _json_bytes(steps_obj))

    return out_path


def import_areplay(store: Store, in_path: str) -> str:
    """
    Imports a .areplay artifact into the store as a NEW run_id.
    Returns the new run_id.
    """
    with zipfile.ZipFile(in_path, "r") as z:
        manifest = json.loads(z.read("manifest.json").decode("utf-8"))
        run_obj = json.loads(z.read("run.json").decode("utf-8"))
        steps_obj = json.loads(z.read("steps.json").decode("utf-8"))

    name = run_obj.get("name", "imported_run")
    started_at = datetime.fromisoformat(run_obj["started_at"])
    meta = run_obj.get("meta", {})
    meta = {**meta, "imported_from": in_path, "areplay_version": manifest.get("version")}

    new_run_id = store.create_run(name=name, started_at=started_at, meta=meta)

    # end_run if ended_at exists
    ended_at = run_obj.get("ended_at")
    if ended_at:
        store.end_run(new_run_id, datetime.fromisoformat(ended_at))

    # Insert steps in order; preserve idx/kind/name/payloads
    for s in sorted(steps_obj, key=lambda x: int(x["idx"])):
        store.add_step(
            run_id=new_run_id,
            idx=int(s["idx"]),
            kind=s["kind"],
            name=s["name"],
            ts=datetime.fromisoformat(s["ts"]),
            input=s.get("input") or {},
            output=s.get("output") or {},
            error=s.get("error"),
        )

    return new_run_id



from __future__ import annotations

import typer
from .store_sqlite import SQLiteStore
from .replay import Replayer
from .diff import diff_runs
from .exporter import export_areplay, import_areplay

app = typer.Typer(no_args_is_help=True)


@app.command("runs")
def runs_cmd(
    action: str = typer.Argument(..., help="list"),
    db: str = typer.Option("agentreplay.db", "--db"),
    limit: int = typer.Option(20, "--limit"),
):
    store = SQLiteStore(db)
    store.init()
    if action != "list":
        raise typer.BadParameter("Only 'list' supported")
    runs = store.list_runs(limit=limit)
    for r in runs:
        typer.echo(f"{r.id}  {r.started_at.isoformat()}  {r.name}")


@app.command("replay")
def replay_cmd(
    run_id: str = typer.Argument(...),
    db: str = typer.Option("agentreplay.db", "--db"),
    strict: bool = typer.Option(False, "--strict", help="Fail if tool inputs differ from recorded trace"),
):
    store = SQLiteStore(db)
    store.init()
    rep = Replayer(store)
    report = rep.replay(run_id, strict=strict)
    typer.echo(f"ok={report.ok} steps={report.steps_replayed}")
    for n in report.notes:
        typer.echo(f"- {n}")


@app.command("diff")
def diff_cmd(
    run_a: str = typer.Argument(...),
    run_b: str = typer.Argument(...),
    db: str = typer.Option("agentreplay.db", "--db"),
):
    store = SQLiteStore(db)
    store.init()
    typer.echo(diff_runs(store, run_a, run_b))


@app.command("export")
def export_cmd(
    run_id: str = typer.Argument(...),
    out: str = typer.Option("run.areplay", "--out", "-o"),
    db: str = typer.Option("agentreplay.db", "--db"),
):
    store = SQLiteStore(db)
    store.init()
    path = export_areplay(store, run_id, out)
    typer.echo(f"Exported: {path}")


@app.command("import")
def import_cmd(
    path: str = typer.Argument(...),
    db: str = typer.Option("agentreplay.db", "--db"),
):
    store = SQLiteStore(db)
    store.init()
    new_id = import_areplay(store, path)
    typer.echo(f"Imported as run_id: {new_id}")


if __name__ == "__main__":
    app()



from __future__ import annotations

import json
from typing import Dict, Any, List

from .store import Store
from .models import ReplayReport, Step


def _stable(obj: Any) -> str:
    return json.dumps(obj, sort_keys=True, ensure_ascii=False)


class ToolMocker:
    def __init__(self, tool_steps: List[Step], strict: bool = False):
        self.strict = strict
        self._tool_steps = tool_steps
        self._cursor = 0

    def next_output(self, tool_name: str, tool_input: Dict[str, Any]) -> Dict[str, Any]:
        if self._cursor >= len(self._tool_steps):
            raise RuntimeError("Replay tool calls exceeded recorded tool calls")

        step = self._tool_steps[self._cursor]
        self._cursor += 1

        if step.name != tool_name:
            raise RuntimeError(
                f"Tool name mismatch at tool-step #{self._cursor-1}: recorded={step.name} current={tool_name}"
            )

        if self.strict:
            if _stable(step.input) != _stable(tool_input):
                raise RuntimeError(
                    f"Tool input mismatch at tool-step #{self._cursor-1} for {tool_name}\n"
                    f"recorded={_stable(step.input)}\ncurrent ={_stable(tool_input)}"
                )

        return step.output


class Replayer:
    def __init__(self, store: Store):
        self.store = store

    def build_tool_mocker(self, run_id: str, strict: bool = False) -> ToolMocker:
        steps = self.store.list_steps(run_id)
        tool_steps = [s for s in steps if s.kind == "tool"]
        return ToolMocker(tool_steps, strict=strict)

    def replay(self, run_id: str, strict: bool = False) -> ReplayReport:
        steps = self.store.list_steps(run_id)
        notes: List[str] = []
        ok = True

        # integrity check
        last_idx = -1
        for s in steps:
            if s.idx != last_idx + 1:
                ok = False
                notes.append(f"Non-contiguous step idx at {s.idx} (prev={last_idx})")
            last_idx = s.idx

        _ = self.build_tool_mocker(run_id, strict=strict)
        notes.append(f"Tool mocker prepared (strict={strict}).")

        return ReplayReport(run_id=run_id, ok=ok, steps_replayed=len(steps), notes=notes)



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, Optional, Iterator

from .store import Store
from .redact import redact_dict


def _utcnow() -> datetime:
    return datetime.now(timezone.utc)


@dataclass
class RunHandle:
    id: str
    name: str
    meta: Dict[str, Any]


class Recorder:
    """
    Framework-agnostic recorder.
    You call recorder.llm(...) / recorder.tool(...) around your agent execution.
    """

    def __init__(self, store: Store, project: str = "default", redaction: bool = True):
        self.store = store
        self.project = project
        self.redaction = redaction
        self._active_run: Optional[RunHandle] = None
        self._idx: int = 0
        self.store.init()

    @contextmanager
    def run(self, name: str, meta: Optional[Dict[str, Any]] = None) -> Iterator[RunHandle]:
        if self._active_run is not None:
            raise RuntimeError("Nested runs not supported in v0.1")
        meta = meta or {}
        meta = {"project": self.project, **meta}

        run_id = self.store.create_run(name=name, started_at=_utcnow(), meta=meta)
        self._active_run = RunHandle(id=run_id, name=name, meta=meta)
        self._idx = 0
        try:
            yield self._active_run
        finally:
            self.store.end_run(run_id, ended_at=_utcnow())
            self._active_run = None
            self._idx = 0

    def _ensure_run(self) -> RunHandle:
        if self._active_run is None:
            raise RuntimeError("No active run. Use `with recorder.run(...):`")
        return self._active_run

    def span(self, name: str, input: Optional[Dict[str, Any]] = None, output: Optional[Dict[str, Any]] = None):
        self._record(kind="span", name=name, input=input or {}, output=output or {}, error=None)

    def llm(self, name: str, input: Dict[str, Any], output: Dict[str, Any], error: Optional[Dict[str, Any]] = None):
        self._record(kind="llm", name=name, input=input, output=output, error=error)

    def tool(self, name: str, input: Dict[str, Any], output: Dict[str, Any], error: Optional[Dict[str, Any]] = None):
        self._record(kind="tool", name=name, input=input, output=output, error=error)

    def _record(self, kind: str, name: str, input: Dict[str, Any], output: Dict[str, Any], error: Optional[Dict[str, Any]]):
        run = self._ensure_run()

        if self.redaction:
            input = redact_dict(input)
            output = redact_dict(output)
            if error is not None:
                error = redact_dict(error)

        idx = self._idx
        self._idx += 1

        self.store.add_step(
            run_id=run.id,
            idx=idx,
            kind=kind,
            name=name,
            ts=_utcnow(),
            input=input,
            output=output,
            error=error,
        )



from __future__ import annotations

import json
import sqlite3
import uuid
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from .models import Run, Step
from .store import Store


def _utcnow() -> datetime:
    return datetime.now(timezone.utc)


def _to_iso(dt: datetime) -> str:
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc).isoformat()


def _from_iso(s: str) -> datetime:
    # Python 3.9 compatible parse for isoformat with timezone
    return datetime.fromisoformat(s)


class SQLiteStore(Store):
    def __init__(self, path: str = "agentreplay.db") -> None:
        self.path = path

    def _conn(self) -> sqlite3.Connection:
        conn = sqlite3.connect(self.path)
        conn.row_factory = sqlite3.Row
        return conn

    def init(self) -> None:
        with self._conn() as c:
            c.execute(
                """
                CREATE TABLE IF NOT EXISTS runs (
                  id TEXT PRIMARY KEY,
                  name TEXT NOT NULL,
                  started_at TEXT NOT NULL,
                  ended_at TEXT NULL,
                  meta_json TEXT NOT NULL
                );
                """
            )
            c.execute(
                """
                CREATE TABLE IF NOT EXISTS steps (
                  id TEXT PRIMARY KEY,
                  run_id TEXT NOT NULL,
                  idx INTEGER NOT NULL,
                  kind TEXT NOT NULL,
                  name TEXT NOT NULL,
                  ts TEXT NOT NULL,
                  input_json TEXT NOT NULL,
                  output_json TEXT NOT NULL,
                  error_json TEXT NULL,
                  FOREIGN KEY(run_id) REFERENCES runs(id)
                );
                """
            )
            c.execute("CREATE INDEX IF NOT EXISTS idx_steps_run ON steps(run_id, idx);")

    def create_run(self, name: str, started_at: datetime, meta: Dict[str, Any]) -> str:
        run_id = str(uuid.uuid4())
        with self._conn() as c:
            c.execute(
                "INSERT INTO runs(id, name, started_at, ended_at, meta_json) VALUES(?,?,?,?,?)",
                (run_id, name, _to_iso(started_at), None, json.dumps(meta or {})),
            )
        return run_id

    def end_run(self, run_id: str, ended_at: datetime) -> None:
        with self._conn() as c:
            c.execute("UPDATE runs SET ended_at=? WHERE id=?", (_to_iso(ended_at), run_id))

    def add_step(
        self,
        run_id: str,
        idx: int,
        kind: str,
        name: str,
        ts: datetime,
        input: Dict[str, Any],
        output: Dict[str, Any],
        error: Optional[Dict[str, Any]],
    ) -> str:
        step_id = str(uuid.uuid4())
        with self._conn() as c:
            c.execute(
                """
                INSERT INTO steps(
                  id, run_id, idx, kind, name, ts, input_json, output_json, error_json
                ) VALUES(?,?,?,?,?,?,?,?,?)
                """,
                (
                    step_id,
                    run_id,
                    idx,
                    kind,
                    name,
                    _to_iso(ts),
                    json.dumps(input or {}),
                    json.dumps(output or {}),
                    None if error is None else json.dumps(error),
                ),
            )
        return step_id

    def get_run(self, run_id: str) -> Run:
        with self._conn() as c:
            row = c.execute("SELECT * FROM runs WHERE id=?", (run_id,)).fetchone()
            if not row:
                raise KeyError(f"Run not found: {run_id}")
            return Run(
                id=row["id"],
                name=row["name"],
                started_at=_from_iso(row["started_at"]),
                ended_at=_from_iso(row["ended_at"]) if row["ended_at"] else None,
                meta=json.loads(row["meta_json"] or "{}"),
            )

    def list_runs(self, limit: int = 50) -> List[Run]:
        with self._conn() as c:
            rows = c.execute(
                "SELECT * FROM runs ORDER BY started_at DESC LIMIT ?", (int(limit),)
            ).fetchall()
            return [
                Run(
                    id=r["id"],
                    name=r["name"],
                    started_at=_from_iso(r["started_at"]),
                    ended_at=_from_iso(r["ended_at"]) if r["ended_at"] else None,
                    meta=json.loads(r["meta_json"] or "{}"),
                )
                for r in rows
            ]

    def list_steps(self, run_id: str) -> List[Step]:
        with self._conn() as c:
            rows = c.execute(
                "SELECT * FROM steps WHERE run_id=? ORDER BY idx ASC", (run_id,)
            ).fetchall()
            return [
                Step(
                    id=r["id"],
                    run_id=r["run_id"],
                    idx=int(r["idx"]),
                    kind=r["kind"],
                    name=r["name"],
                    ts=_from_iso(r["ts"]),
                    input=json.loads(r["input_json"] or "{}"),
                    output=json.loads(r["output_json"] or "{}"),
                    error=json.loads(r["error_json"]) if r["error_json"] else None,
                )
                for r in rows
            ]



from __future__ import annotations

import json
from typing import Any, List, Tuple, Optional
from .store import Store
from .models import Step


def _stable(obj: Any) -> str:
    return json.dumps(obj, sort_keys=True, ensure_ascii=False)


def _first_divergence(a_steps: List[Step], b_steps: List[Step]) -> Optional[Tuple[int, Step, Step, str]]:
    n = min(len(a_steps), len(b_steps))
    for i in range(n):
        a = a_steps[i]
        b = b_steps[i]
        if (a.kind, a.name) != (b.kind, b.name):
            return (i, a, b, "kind/name changed")
        if _stable(a.input) != _stable(b.input):
            return (i, a, b, "input changed")
        if _stable(a.output) != _stable(b.output):
            return (i, a, b, "output changed")
        if _stable(a.error) != _stable(b.error):
            return (i, a, b, "error changed")
    if len(a_steps) != len(b_steps):
        # Divergence at length boundary
        i = n
        dummy = a_steps[-1] if a_steps else b_steps[-1]
        return (i, dummy, dummy, f"length changed: {len(a_steps)} != {len(b_steps)}")
    return None


def diff_runs(store: Store, run_a: str, run_b: str) -> str:
    a_steps = store.list_steps(run_a)
    b_steps = store.list_steps(run_b)

    div = _first_divergence(a_steps, b_steps)
    if div is None:
        return "No divergence. Runs are identical at step granularity."

    i, a, b, reason = div
    lines = []
    lines.append(f"DIVERGENCE at step index {i}: {reason}")
    lines.append(f"  A: kind={a.kind} name={a.name} idx={a.idx}")
    lines.append(f"  B: kind={b.kind} name={b.name} idx={b.idx}")

    # Show small payload snippets
    lines.append("  A.input:  " + _stable(a.input)[:400])
    lines.append("  B.input:  " + _stable(b.input)[:400])
    lines.append("  A.output: " + _stable(a.output)[:400])
    lines.append("  B.output: " + _stable(b.output)[:400])
    return "\n".join(lines)
